/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package electro.core;

import electro.entities.Coalition;
import electro.entities.PowerUser;
import electro.entities.UsersPeakConsumptionReverseComparator;
import electro.entities.Value;
import electro.entities.ValueReverseComparator;
import electro.entities.ValueSet;
import electro.view.MainWindow;
import electro.exceptions.TimeIntervalsNumberMismatchException;
import electro.exceptions.FinishedException;
import java.util.ArrayList;
import java.util.BitSet;

/**
 * This class implements the actual optimization algorithm
 * @author Andrea Zambon
 */
public class Coalizer implements Runnable {

    private static final int PROGRESS_BAR_VARIABLE_NUMBER = 6;

    private int n;                           //The number of users
    private PowerUser[] u;                   //Array of users
    private Coalition[] coalitions;          //Array of coalitions

    //The current variable order. The value at position 0 is the variable at
    //level 0, the value at position 1 is the variable at level 1 and so on
    private int[] variableOrder;

    //Current variable assignments. assignments[i] is the value of Xi. Notice
    //that the variables are not instantiated in order, so, for example, X4 may
    //be instantiated before X1
    private int[] assignments;

    //Bit set representing which variables are already assigned. A value of 1 at
    //position i means that Xi is already instantiated
    private BitSet assignedVariables;

    //Array of bit sets representing the forbidden values for each variable.
    //A value of 1 in the j-th position of the i-th bit set means that the
    //assignment Xi = j is forbidden.
    private BitSet[] forbiddenValues;

    //This array of bit sets represents the forbidden values added at each level.
    //it is used to efficiently update the forbiddenValues array during
    //backtrack and value change operations. Since an assignment of a variable
    //Xi can only prevent other variables from assuming the same value as Xi,
    //this array of bit sets only stores which variables had the value of Xi
    //forbidden. A value of 1 in the j-th position of the i-th bit set means
    //that at the i-th level the value of the current variable became forbidden
    //for the variable Xj. In other words, if forbiddenValuesAtLevel[i].get(j) == true,
    //variableOrder[i] == k and assignments[k] == val, then at level i the value
    //val became forbidden for variable Xj due to the assignment Xk = val.
    private BitSet[] forbiddenValuesAtLevel;

    //Current backtracking level. The root of the tree is level 0 and the leaves
    //are level n-1
    private int level;

    //Index of the current variable. Shortcut for variableOrder[level]
    private int currentVariableIndex;

    //Value of the current variable. Shortcut for assignment[currentVariableIndex]
    private int currentVariableValue;

    //The index of the current value of each variable
    private int[] currentVariableValueIndices;

    //Current lower bound
    private float lowerBound;

    //Array containing the maximum values for the variable at each level. This
    //is the maximum of the values assigned to all other variables in the order
    //plus one. This maximum only depends on the value of higher priority
    //variables, so it is only computed whenever the algorithm moves down the
    //tree from a node to one of its children. It can be computed quickly in
    //constant time: The maximum for Xi is equal to the maximum for its father
    //Xj except when the value of the father is equal to its maximum. In this
    //case the maximum for Xi is the maximum for Xj plus one. In addition any
    //maximum cannot be greater than the highest valid index of a coalition.
    //So MaxValue(Xi) = min(maxCoalitionIndex, max(MaxValue(Xj), Xj + 1))
    private int[] maxValueAllowedAtLevel;

    //This array is used to contain the best assignment found so far. The value
    //at position i is the value for Xi.
    private int[] bestSolution;

    //Window object used to control the screen output
    private MainWindow window;

    private ArrayList<PowerUser> users;

    private float coalitionSize;

    private int coalitionNumber;

    private boolean stop;

    private int threadCount;

    private ValueSet[] values;

    private ValueReverseComparator valueReverseComparator;

    private float[] minUserConsumption;
    private float[] maxUserConsumption;

    private float[] coalitionsMaxComsumptions;

    /**
     * Creates a new coalizer object
     *
     * @param usersArray An array of users that has to be split into a series of
     * coalitions
     *
     * @param coalitionSizeParam The maximum cumulative power consumption of
     * each coalition
     *
     * @param coalitionNumberParam The maximum number of coalitions thet can be
     * used by the algorithm
     *
     * @param w MainWindow object that will be used to display the solution
     * found
     * 
     * @throws TimeIntervalsNumberMismatchException Thrown if not all users have
     * the same number of time intervals
     *
     * his constructor creates a coalizer object and prepares it to start the
     * computation o find the best grouping of the provided users array into
     * a series of coalitions.
     */
    public Coalizer(ArrayList<PowerUser> usersArray, float coalitionSizeParam,
                int coalitionNumberParam, MainWindow w)
                throws TimeIntervalsNumberMismatchException {

        //Save the window object to show updated solutions
        window = w;

        //Create an empty array of users
        users = new ArrayList<PowerUser>();
        
        //Fill the array with deep copies of each user
        for (int i = 0; i < usersArray.size(); ++i) {

            users.add(usersArray.get(i).clone());
        }

        //Create an array of floats to avoid having to query every coalition
        //for its peak cumulative power consumption over and over
        coalitionsMaxComsumptions = new float[coalitionNumberParam];

        //Save the coalition size
        coalitionSize = coalitionSizeParam;

        //Save the maximum number of coalitions allowed
        coalitionNumber = coalitionNumberParam;

        //Try initializing the algorithm to catch potential exceptions
        initialize();

        stop = true;

        //No thread running
        threadCount = 0;

        valueReverseComparator = new ValueReverseComparator();
    }

    /**
     * Initializes all internal parameters before the main algorithm can
     * begin its execution.
     *
     * @throws TimeIntervalsNumberMismatchException Thrown if not all users have
     * the same number of time intervals
     *
     * This method is called before tha main computation loop starts. It
     * initialzes several internal daa structures and variables used by the main
     * algorithm.
     */
    private void initialize() throws TimeIntervalsNumberMismatchException {

        //Get the number of users
        n = users.size();

        //Handle the case of zero users
        if (n == 0) {

            //Stop here
            return;
        }
        
        //Get the number of time intervals for the users
        int timeIntervals = users.get(0).timeIntervalNumber();

        
        
        //Create an array to contain all possible coalitions that will be filled
        coalitions = new Coalition[coalitionNumber];

        //Create all the coalitions
        for (int i = 0; i < coalitionNumber; ++i) {

            //Create a new coalition
            coalitions[i] = new Coalition(coalitionSize, timeIntervals, n);
        }
        
        //Create a standard array to contain a local copy of the users
        u = new PowerUser[n];

        //Create the arrays that will hold the minimum and maximum power usage
        //of each user
        minUserConsumption = new float[n];
        maxUserConsumption = new float[n];

        //Initialize the lower bound to a value equal to the maximum number of
        //coalitions minus one. This is to prevent a problem if the only possible
        //solution is the one that places every user in a separate coalition.
        //With lowerBound == coalitionNumber in such a case the upper bound for
        //the only correct solution may not be greater than the lower bound, and
        //so the solution would be skipped.
        lowerBound = coalitionNumber - 1;

        for (int i = 0; i < n; ++i) {

            //Copy the next user into the array
            u[i] = users.get(i);

            //Check the number of time intervals for this user
            if (u[i].timeIntervalNumber() != timeIntervals) {

                //Error: all users must have their power consumption data
                //represented with the same number of time intervals
                throw new TimeIntervalsNumberMismatchException();
            }

            minUserConsumption[i] = u[i].getMinConsumption();
            maxUserConsumption[i] = u[i].getMaxConsumption();
        }

        //Sort the users according to their peak energy consumption in
        //descending order
        java.util.Arrays.sort(u, new UsersPeakConsumptionReverseComparator());

        //Initialize all remaining staic fields of this class
        assignments                 = new int[n];
        forbiddenValues             = new BitSet[n];
        forbiddenValuesAtLevel      = new BitSet[n];

        bestSolution                = new int[n];

        //Just set the first element to represent that no solution has been
        //found yet
        bestSolution[0]             = -1;

        for (int i = 0; i < n; ++i) {

            forbiddenValues[i] = new BitSet();

            //No variable is assigned initially
            assignments[i]  = -1;

            //The index is the current level. It represents all values of
            //unassigned variables that conflict with the newly assigned variable
            //at level i. Basically it is the set of unassigned variables that
            //before the last assignment could have the value of the last variable
            //assigned to them, and that now they cannot.
            forbiddenValuesAtLevel[i] = new BitSet();
        }

        //Set this to the same value as each element in the assignments array
        currentVariableValue = -1;

        assignedVariables = new BitSet();

        //Max value of each variable
        maxValueAllowedAtLevel = new int[n];

        //Variable order. Can change during execution
        variableOrder = new int[n];

        //Set the first variable
        currentVariableIndex = 0;

        //Mark the first variable as assigned
        assignedVariables.set(currentVariableIndex);

        //save the index of the first variable
        variableOrder[0] = currentVariableIndex;

        //The initial level is 0
        level = 0;

        //Only value 0 allowed for the first variable
        maxValueAllowedAtLevel[0] = 0;

        //Allocate all the ValueSet objects
        values = new ValueSet[n];

        //For each value set, allocate its internal array
        for (int i = 0; i < n; ++i) {

            values[i] = new ValueSet();
            
            values[i].values = new Value[coalitionNumber];
            values[i].valueNumber = 0;

            for (int j = 0; j < coalitionNumber; ++j) {

                values[i].values[j] = new Value();
            }
        }

        currentVariableValueIndices = new int[n];

        //Initialize the index of the last value checked for the first variable
        currentVariableValueIndices[0] = -1;

        //Initialize the valid values for the first variable
        if (coalitions[0].canBeAdded(u[0])) {

            //Initialize the values for the forst variable. No need to compute
            //the upper bound, as there is only one value
            values[0].valueNumber = 1;
            values[0].values[0].value = 0;
            values[0].values[0].ub = Float.POSITIVE_INFINITY;
        }
        else {

            //Let the algorithm fail on its own. Remember that no value were
            //found
            values[0].valueNumber = 0;
        }

        //Call the garbage collector before beginning the main computation loop
        System.gc();

        //Done
        return;
    }

    /**
     * Main computation loop
     *
     * @return An array of coalitions representing the best grouping of users or
     * null if no such grouping exists.
     *
     * @throws TimeIntervalsNumberMismatchException Thrown if not all users have
     * the same number of time intervals
     *
     * This method contains the main computation loop of the program.
     */
    private ArrayList<Coalition> coalize() throws TimeIntervalsNumberMismatchException {

        //Make sure that there is at least one user
        if (users.isEmpty()) {

            //No users. Trivial solution
            return new ArrayList<Coalition>();
        }

        //Create an empty vector to store the result
        ArrayList<Coalition> result = new ArrayList<Coalition>();

        //Initialize all data structures
        initialize();

        try {

            //Main loop
            while (!stop) {

                //Choose a value for the current variable or backtrack if necessary.
                //If the end of the search is reached, an exception is thrown
                chooseNewValueOrBacktrack();

                //Propagate all constraints as much as possible
                forwardCheck();

                //Advance to the next variable or update the lower bound if
                //there is no next variable (if this is a complete assignment).
                chooseNextVariableToInstantiate();
            }
        }

        catch (FinishedException ex) {

            //Do nothing. The exception is only needed to leave the loop
        }

        //Was a solution found?
        if (bestSolution[0] == -1) {

            //No solution found
            result = null;
        }
        else {

            //Remove anything that may be in the coalitions (if the algorithm
            //was stopped before it completed)
            for (int i = 0; i < coalitionNumber; ++i) {
                
                coalitions[i].clear();
            }

            //Recreate the best solution found
            for (int i = 0; i < n; ++i) {

                coalitions[bestSolution[i]].push(u[i]);
            }

            //Add to the result all nonempty coalitions
            for (int i = 0; i < coalitionNumber; ++i) {

                //Check the number of users in this coalition
                if (coalitions[i].numberOfUsers() > 0) {

                    //Not an empty coalition. Add this to the result
                    result.add(coalitions[i]);
                }
            }
        }

        //Cleanup
        n = 0;
        u = null;
        coalitions = null;

        variableOrder = null;
        assignments = null;
        assignedVariables = null;
        forbiddenValues = null;
        forbiddenValuesAtLevel = null;

        level = 0;
        currentVariableIndex = 0;
        currentVariableValue = 0;
        lowerBound = 0;

        maxValueAllowedAtLevel = null;
        bestSolution = null;

        //Call the garbage collector to clean up as much memory as possible
        System.gc();

        //Leave the stop variable set to true
        stop = true;

        //Return the best result found
        return result;
    }

    /**
     * Chooses a new value for a variable
     *
     * @throws FinishedException Thrown if the computation is finished
     *
     * This method tries to find the next value for the last assigned variable.
     * It tries several values unil one is found that does not violate any
     * constraint and whose upper bound is better than the current lower bound
     * or all values have been searched unsuccesfully. In the first case the
     * method returns, while in the second case the method performs a backtrack
     * to the previous variable and the process is repeated.
     */
    private void chooseNewValueOrBacktrack() throws FinishedException {

        int newValue;

        while (true) {

            //If there was a previous value assigned to this variable, clear
            //all the information generated by such an assignment
            if (currentVariableValue != -1) {

                //Remove the user from the coalition it is in
                coalitions[currentVariableValue].pop();

                //Remove any forbidden value that was computed at this level
                restoreForbiddenValuesAtThisLevel();
            }

            //Get the index of the next value
            int valueIndex = currentVariableValueIndices[currentVariableIndex] + 1;

            if (valueIndex < values[currentVariableIndex].valueNumber &&
                    values[currentVariableIndex].values[valueIndex].ub > lowerBound) {
                
                //Extract the next value for this variable
                newValue = values[currentVariableIndex].values[valueIndex].value;

                //Put this value in the assignment
                assignments[currentVariableIndex] = newValue;

                try {

                    //Try put this user in the specified coalition
                    coalitions[newValue].push(u[currentVariableIndex]);
                }
                
                catch (TimeIntervalsNumberMismatchException ex) {
                    
                    //The initialization phase should catch all these exceptions,
                    //so it should never be thrown here
                }

                //Update the assignment
                currentVariableValue = newValue;

                if (level == PROGRESS_BAR_VARIABLE_NUMBER) {

                    //Update the progress bar
                    float progress = 0;
                    float section = 1;

                    //Do not consider the first variable, since it only has
                    //a single possible value
                    for (int i = 1; i <= PROGRESS_BAR_VARIABLE_NUMBER; ++i) {

                        //Compute the new value for the progress bar
                        progress +=
                                (currentVariableValueIndices[variableOrder[i]] * section)
                                / (float) (values[variableOrder[i]].valueNumber);

                        //Compute the section of bar affected by the next
                        //variable value
                        section /= (float) (values[variableOrder[i]].valueNumber);
                    }

                    //Convert the value to a percentage and apply it to the
                    //bar on the window
                    if (window != null) {
                    
                        window.updateProgressBar((int) (progress * 100));
                    }
                }

                //Save the current index of this variable
                currentVariableValueIndices[currentVariableIndex] = valueIndex;

                //Done
                return;
            }
            
            //If we arrive here, it was not possible to find a new value for the
            //current variable. So it's necessary to backtrack and repeat
            backtrack();
        }
    }

    /**
     * Backtracks to the previous variable
     *
     * @throws FinishedException Thrown if the computation is finished
     *
     * This method performs a backtrack to the previous variable. If there is no
     * previous variable, then the algorithm is teminated as no other value has
     * to be searched.
     */
    private void backtrack() throws FinishedException {

        //Are we backtracking from the top variable?
        if (level == 0) {

            //Finished. Raise an exception to leave the main loop
            throw new FinishedException();
        }

        //Mark this variable as not assigned
        assignedVariables.clear(currentVariableIndex);

        //Reset the value of this variable to the special value -1
        assignments[currentVariableIndex] = -1;

        //Move back to the previous level
        --level;

        //Get the index of the new current variable
        currentVariableIndex = variableOrder[level];
        
        //Get the value of the new current variable
        currentVariableValue = assignments[currentVariableIndex];
    }

    /**
     * Propagates forbidden values to unassigned variables
     *
     * This method computes the forbidden values for each unassigned variable
     * due to the last variable that was assigned.
     */
    private void forwardCheck() {

        //Loop over all unassigned variables
        for (int variableIndex = 0; variableIndex < n; ++variableIndex) {

            //Is this variable unassigned?
            if (!assignedVariables.get(variableIndex)) {

                try {
                    
                    //Unassigned variable. Can it be added to the coalition
                    //that the current variable belongs to?
                    if (!forbiddenValues[variableIndex].get(currentVariableValue) &&
                            !coalitions[currentVariableValue].canBeAdded(u[variableIndex])) {

                        //Cannot be added. Mark the value of the current
                        //variable as forbidden for this variable
                        forbiddenValues[variableIndex].set(currentVariableValue);

                        //Remember that this forbidden value was added at this
                        //level
                        forbiddenValuesAtLevel[level].set(variableIndex);
                    }
                }

                catch (TimeIntervalsNumberMismatchException ex) {

                    //This was already checked in the initialization and should
                    //not happen here
                }
            }
        }
    }

    /**
     * Restores any value that was forbidden by a variable at this level
     *
     * This method restores all the values that were previously forbidden due to
     * the value of the variable that is at the current level.
     */
    private void restoreForbiddenValuesAtThisLevel() {

        //Is there something to do?
        if (!forbiddenValuesAtLevel[level].isEmpty()) {

            //cleared is the counter of bits cleared
            int cleared = 0;

            //variable is the current variable index to clear
            int variable = 0;

            //Get the total number of bits to clear
            final int totalToClear = forbiddenValuesAtLevel[level].cardinality();

            //Repeat until all values have been cleared
            while (cleared < totalToClear) {

                //Should we clear the bit for this variable?
                if (forbiddenValuesAtLevel[level].get(variable)) {

                    //Restore the value of the last assigned variable to
                    //the domain of all variables that had such value removed
                    forbiddenValues[variable].clear(currentVariableValue);

                    //One more value cleared
                    ++cleared;
                }

                //Move to the next variable
                ++variable;
            }

            //Reset all bits to false at this level
            forbiddenValuesAtLevel[level].clear();
        }

        //Done
        return;
    }

    /**
     * Chooses the next variable that should be instantiated and detects
     * the presence of a complete solution
     *
     * This method tries to find the best variable to assign next to move toward
     * a solution of the problem. The variable is chosen according to the most
     * constrained variable heuristic. If no unassigned vaiable exists, then
     * this method detects it and checks the complete solution found. If
     * necessary the current lower bound is updated and the best solution found
     * is saved.
     */
    private void chooseNextVariableToInstantiate() {

        //Remember the index ond the number of forbidden values for the
        //unassigned variable with the most forbidden values.
        int best;
        int bestForbiddenValues;
        
        //Find the first unassigned variable
        for (int i = 0; i < n; ++i) {
            
            //Is this variable unassigned?
            if (!assignedVariables.get(i)) {
                
                //Save the index and the number of forbidden values for this
                //variable
                best = i;
                bestForbiddenValues = forbiddenValues[i].cardinality();

                //Is there another better variable?
                for (++i; i < n; ++i) {

                    //Is this unassigned?
                    if (!assignedVariables.get(i)) {

                        //Does this have more forbidden valus than the current
                        //best?
                        if (forbiddenValues[i].cardinality() > bestForbiddenValues) {
                            
                            //New best variable
                            best = i;
                            bestForbiddenValues = forbiddenValues[i].cardinality();
                        }
                    }
                }

                //Advance to the next level
                ++level;

                //Save the index of the new current variable
                variableOrder[level] = best;

                //Update the index and value of the current variable
                currentVariableIndex = best;
                currentVariableValue = -1;

                //Compute the maximum between the maximum of the previous level
                //and the value of the variable at the previous level plus one.
                //Then make sure that the result does not become greater than the
                //maximum coalition index
                maxValueAllowedAtLevel[level] = java.lang.Math.min(
                        coalitions.length - 1,
                        java.lang.Math.max(maxValueAllowedAtLevel[level - 1],
                                        assignments[variableOrder[level - 1]] + 1));

                //Mark this variable as assigned
                assignedVariables.set(currentVariableIndex);

                //Remember the number of valid values found
                int counter = 0;

                //Find all the values for the next variable
                for (int j = 0; j <= maxValueAllowedAtLevel[level]; ++j) {

                    //Only consider non forbidden values
                    if (!forbiddenValues[currentVariableIndex].get(j)) {

                        //Value j is not forbidden
                        try {

                            //Temporarily add the current user to the j-th coalition
                            coalitions[j].push(u[currentVariableIndex]);

                            //Compute and save the upper bound for this value
                            values[currentVariableIndex].values[counter].ub = upperBound();
                            values[currentVariableIndex].values[counter].value = j;

                            //Increment the number of values found if this value
                            //is better than the current lower bound
                            if (values[currentVariableIndex].values[counter].ub > lowerBound) {

                                ++counter;
                            }

                            //Restore the previous situation by removing the user
                            //from the coalition
                            coalitions[j].pop();
                        }

                        catch (TimeIntervalsNumberMismatchException ex) {

                            //Should not happen here
                        }
                    }
                }

                //Save the number of values found
                values[currentVariableIndex].valueNumber = counter;

                //Now sort the values in descending order according to their
                //upper bound. Only sort the used part of the array
                java.util.Arrays.sort(values[currentVariableIndex].values,
                                0, values[currentVariableIndex].valueNumber,
                                valueReverseComparator);

                //Set the index of the last value used for this variable to -1.
                //So in the next iteration the algorithm will start from index 0
                currentVariableValueIndices[currentVariableIndex] = -1;

                //Done
                return;
            }
        }

        //No variable is unassigned, that is the assignment is complete. Compute
        //the value of this assignment.
        float solutionValue = computeValue();

        //Check and update the lower bound if necessary
        if (solutionValue > lowerBound) {

            //Better than the previous lower bound. Update the lower bound
            lowerBound = solutionValue;

            //Save the solution
            for (int i = 0; i < n; ++i) {

                bestSolution[i] = assignments[i];
            }

            //Show the current best solution
            if (window != null) {
            
                window.showSolution(coalitions);
            }
        }

        //Nothing else to do here. The main loop does not know that a complete
        //assignment was found and that the algorithm did not move to another
        //variable. In the next iteration of the loop, a new value will be
        //searched for the last variable, either finding another complete
        //assignment or causing the procedure to backtrack to some other
        //previous variable.
        return;
    }

    /**
     * This method computes an upper bound to the best possible solution
     * containing the current partial variable assignment
     *
     * @return An upper bount to the best possible value of a solution that
     * contains the current partial assignment.
     *
     * This method computes the upper bound to the best possible value that can
     * be obtained by a solution containing the current partial assignment. This
     * method is the single main reason for the efficiency or inefficiency of
     * the algorithm: if this bound is very tight, then the algorithm will cut
     * much of the search space and find the optimum solution very quickly.
     * Otherwise, if the bound is very loose, the algorithm degenerates to a
     * simple backtracking search.
     */
    private float upperBound() {

        //Compute the objective function using only assigned variables
        float ub = 0;

        for (int i = 0; i < coalitionNumber; ++i) {

            //Compute the LDF for this coalition and add it to the result
            ub += coalitions[i].LDF();

            //Save the peak cumulative power consumption of each coalition
            coalitionsMaxComsumptions[i] = coalitions[i].getMaxConsumption();
        }

        float minMax = Float.POSITIVE_INFINITY;
        int number = 0;

        //Find the smallest maximum power consumption of any unassigned user.
        //Also count the number of unassigned variables
        for (int i = 0; i < n; ++i) {

            if (!assignedVariables.get(i)) {

                ++number;

                if (maxUserConsumption[i] < minMax) {

                    //New minimum maximum power consumption
                    minMax = maxUserConsumption[i];
                }
            }
        }

        //Shortcut: if no value was found, it means that all the users are
        //already assigned, and so the upper bound is actually the exact value
        //of the objectuve function
        if (minMax == Float.POSITIVE_INFINITY) {

            //Just return the value computed by the initial loop
            return ub;
        }

        float sumOfIncrements = 0;

        //For each variable find he coalitions with the smallest peak cumulative
        //power consumption that is not forbidden
        float min;
        boolean found;

        //Iterate on all variables
        for (int i = 0; i < n; ++i) {

            //Only consider unassigned variables
            if (!assignedVariables.get(i)) {

                //Unassigned variable
                found = false;
                min = Float.POSITIVE_INFINITY;

                //Find the coalition with the smallest peak cumulative power
                //consumption that is not forbidden for this variable
                for (int j = 0; j < coalitionNumber; ++j) {

                    //Is this forbidden?
                    if (!forbiddenValues[i].get(j)) {

                        //Not a forbidden value. Remember that at least an
                        //admissible value was found for this variable
                        found = true;

                        //Is this an empty coalition?
                        if (coalitions[j].numberOfUsers() == 0) {

                            //Empty coalition. Consider the best situation that
                            //can be obtained, that is putting in this coalition
                            //only the user with the smallest peak power
                            //consumption
                            min = minMax;
                        }
                        else {

                            //Get the peak cumulative power consumption for this
                            //coalition
                            min = coalitionsMaxComsumptions[j];
                        }

                        //Look for other non-forbidden values
                        for (++j; j < coalitionNumber; ++j) {

                            //Is this coalition forbidden?
                            if (!forbiddenValues[i].get(j)) {

                                //Not forbidden. Is this empty?
                                if (coalitions[j].numberOfUsers() == 0) {

                                    //Empty coalition Consider the best
                                    //situation that can be obtained, that is
                                    //putting in this coalition only the user
                                    //with the smallest peak power consumption
                                    if (minMax < min) {

                                        min = minMax;
                                    }
                                }
                                else {

                                    //Not an empty coalition. If necessary
                                    //update the minimum peak cumulative power
                                    //consumption of a non forbidden coalition
                                    if (coalitionsMaxComsumptions[j] < min) {

                                        //Update
                                        min = coalitionsMaxComsumptions[j];
                                    }
                                }
                            }
                        }
                        //All coalitions checked

                        //Leave the outer loop as well
                        break;
                    }
                }

                if (!found) {

                    //No allowed value for this variable. This partial assignment
                    //cannot be extended to form a complete solution
                    return -1;
                }

                //Compute the maximum possible increment that this variable can
                //produce and accumulate it.
                sumOfIncrements +=
                    (maxUserConsumption[i] - minUserConsumption[i]) /
                    (Math.max(min + minUserConsumption[i], maxUserConsumption[i]));
            }
        }

        //Compute the new upper bound and return it
        return ub + sumOfIncrements;
    }

    /**
     * Computes the value for a complete solution
     *
     * @return The value for a given complete solution
     *
     * This method computes the value for a complete solution of the problem
     */
    private float computeValue() {

        //Compute the exact value of the objective function for a complete
        //assignment. Just sum the LDF of each coalition
        float ub = 0;

        for (int i = 0; i < coalitionNumber; ++i) {

            //Compute the LDF for this coalition and add it to the result
            ub += coalitions[i].LDF();
        }

        return ub;
    }

    /**
     * This method is run by any newly created thread in this object
     *
     * This method is used to implement the Runnable interface. It is run by any
     * thread that is started in this object.
     */
    public void run() {

        //Initialize the solution for this thread to null
        ArrayList<Coalition> solution = null;

        //Set the flag to represent whether this thread executed the algorithm
        //to false
        boolean executed = false;

        synchronized(this) {

            //Only one thread allowed at a time
            if (threadCount == 0) {

                //Increase the number of threads
                ++threadCount;
            }
            else {

                //Another thread already running. Stop here
                return;
            }
        }

        try {

            //Do not stop the thread immediately
            stop = false;

            //Start the main computation
            solution = coalize();

            //Remember that this thread executed the algorithm
            executed = true;
        }
        catch (TimeIntervalsNumberMismatchException ex) {
            
            //Should never happen here. The operation generating this was
            //already executed in the constructor
        }

        synchronized(this) {

            //Reduce the number fo threads running
            --threadCount;
        }

        //Make sure that this thread did indeed execute the algorithm and didn't
        //arrive here due to an exception
        if (executed) {

            //Save the solution found
            if (window != null) {
            
                window.saveSolution(solution);

                //Compute the resulting LDF
                float LDF = 0;

                int i = 0;

                //Nonempty coalitions
                if (solution != null) {
                    
                    for (i = 0; i < solution.size(); ++i) {

                        LDF += solution.get(i).LDF();
                    }
                }

                //Empty coalitions
                for (; i < coalitionNumber; ++i) {

                    LDF += 1;
                }

                //Write the result
                System.out.println("Optimum LDF: " + LDF);
            }
        }

        //Signal that the algorithm finished its execution
        if (window != null) {
        
            window.done();
        }
    }

    /**
     * Creates a new thread to run the algorithm in this object
     *
     * This method creates a new thread to run the algorithm in this object and
     * starts it.
     */
    public void start() {

        //Create a new thread
        Thread t = new Thread(this);

        //Go
        t.start();
    }

    /**
     * Stops the thread executing the algorithm
     *
     * This method signals the running thread that it should stop as soon as
     * possible
     */
    public void stop() {

        //Signal the core thread to stop as soon as possible
        stop = true;
    }
}
