/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package electro.entities;

import electro.exceptions.TimeIntervalsNumberMismatchException;
import java.util.EmptyStackException;

/**
 * This class represents a coalition of users
 *
 * @author Andrea Zambon
 */
public class Coalition {

    private PowerUser[] users;
    private int usersNumber;

    private final float maxConsumptionAllowed;

    private float[][] cumulative;
    private float[] currentMaxConsumption;

    private final int intervals;
    private final int maxUsersNumber;

    private float[] sumOfMaximums;

    /**
     * Creates a new empty coalition
     *
     * @param max The maximum cumulative power consumption value for this
     * coalition
     *
     * @param timeIntervals he number of time intervals used to represent the
     * power consumption profile of each user that will be added to this
     * coalition
     *
     * @param maxUsers Maximum number of users that may be added to this
     * coalition
     *
     * This constructor creates an empty coalition with the specified maximum
     * aggregate power consumption, number of time intervals used to represent
     * the cumulatove power consumption and the specified maximum number of
     * users.
     */
    public Coalition(float max, int timeIntervals, int maxUsers) {

        //Create the arrays
        users = new PowerUser[maxUsers];
        cumulative = new float[maxUsers + 1][timeIntervals];
        currentMaxConsumption = new float[maxUsers + 1];
        sumOfMaximums = new float[maxUsers + 1];
        
        //Initial number of users
        usersNumber = 0;

        //Max allowed power consumption
        maxConsumptionAllowed = max;

        //Zero out the initial cumulative power consumption
        for (int i = 0; i < timeIntervals; ++i) {

            cumulative[0][i] = 0.0f;
        }

        //Save the number of time intervals
        intervals = timeIntervals;

        //Save the maximum number of users that can be saved in the array
        maxUsersNumber = maxUsers;

        //Initialize the maximum power consumption
        currentMaxConsumption[0] = 0;

        //Initialize the sum of maximum user power demands
        sumOfMaximums[0] = 0.0f;
    }

    /**
     * Returns the number of users currently present
     *
     * @return The number of users currently present
     *
     * This method returns the number of users currently in this coalition
     */
    public final int numberOfUsers() {

        return usersNumber;
    }

    /**
     * Returns a reference to the index-th user
     *
     * @param index The index of the user whose reference is to be returned
     *
     * @return A reference to the index-th user in this coalition
     *
     * @throws ArrayIndexOutOfBoundsException Thrown if the index is not valid
     *
     * This method returns a reference to the index-th user in this coalition
     */
    public PowerUser getUser(int index) {

        //Make sure that the index is valid
        if (index < 0 || index >= usersNumber) {

            throw new ArrayIndexOutOfBoundsException();
        }

        return users[index];
    }

    /**
     * This method returns the maximum cumulative power consumption
     * allowed in this coalition.
     *
     * @return The maximum allowed power consumption in this coalition
     *
     * This method returns the maximum cumulative power consumption allowed in
     * this coalition
     */
    public final float getMaxConsumptionAllowed() {

        return maxConsumptionAllowed;
    }

    /**
     * Returns the current maximum cumulative power consumption for this
     * coalition
     *
     * @return The current maximum cumulative power consumption for this coalition
     *
     * This method returns the current maximum cumulative power consumption for this
     * coalition
     */
    public final float getMaxConsumption() {
        
        return currentMaxConsumption[usersNumber];
    }

    /**
     * This method returns the sum of the maximums of the power consumption
     * of each use in this coalition
     *
     * @return The sum of the maximums of the power consumption
     * of each use in this coalition
     *
     * This method returns the sum of the maximums of the power consumption
     * of each use in this coalition
     */
    public final float getSumOfMaximums() {

        return sumOfMaximums[usersNumber];
    }

    /**
     * Compute the load diversity factor for this coalition
     *
     * @return The load diversity factor for this coalition
     *
     * This method computes and returns the load diversity factor for this
     * coalition. The load diversity factor is defined as the sum of the maximum
     * power consumption of each user divided by the maximum of the cumulative
     * power consumption. If no user is present, the default value of 1 is used.
     */
    public final float LDF() {

        if (currentMaxConsumption[usersNumber] > 0) {

            //Definition of load diversity factor
            return sumOfMaximums[usersNumber] / currentMaxConsumption[usersNumber];
        }
        else {

            //Default value of 1 when no user is in this coalition
            return 1;
        }
    }

    /**
     * Removes all the users from this coalition
     *
     * This method removes all the users from this coalition
     */
    public final void clear() {

        //It's not necessary to clear the arrays. Just ignore everything that
        //may be present in them

        //Reset the number of users
        usersNumber = 0;
    }

    /**
     * Tests whether a given user can be added to this coalition without
     * passing the maximum allowed cumulative power consumption value
     *
     * @param u The new user to be tested
     *
     * @return True if the new user can be added without passin the maximum
     * allowed cumulative power consumption, false otherwise
     *
     * @throws TimeIntervalsNumberMismatchException Thrown if the number of time
     * intervals used to describe the power consumption profile of the new user
     * does not equal the number of time intervals used by this coalition.
     *
     * This method is used to test whether a given user can be added to this
     * coalition without passing the maximum allowed cumulative power consumption
     * value.
     */
    public final boolean canBeAdded(PowerUser u) throws TimeIntervalsNumberMismatchException {

        //Check that the number of time intervals is the same
        if (u.timeIntervalNumber() != intervals) {

            //Error: the number of time intervals must match
            throw new TimeIntervalsNumberMismatchException();
        }

        //Get a local copy of the power usage of this user
        float[] usage = u.getUsages();

        for (int i = 0; i < intervals; ++i) {

            //Does this exceed the maximum power consumption allowed?
            if (cumulative[usersNumber][i] + usage[i] > maxConsumptionAllowed) {

                //Too high
                return false;
            }
        }

        //Everything's right
        return true;
    }

    /**
     * Checks whether the maximum allowed cumulative power consumption is
     * passed or not
     *
     * @return True if the maximum cumulative power consumption is not above the
     * maximum allowed value, false otherwise.
     *
     * This method tests whether this coalition is valid, that is whether the
     * cumulative power conumption is not greater than the maximum allowed value
     */
    public boolean isAValidCoalition() {

        //Make sure that the current total power consumption does not exceed the
        //maximum allowed power consumption
        return currentMaxConsumption[usersNumber] <= maxConsumptionAllowed;
    }

    /**
     *  Adds a new user at the and of this coaliton
     * @param u The new user to add
     * @throws TimeIntervalsNumberMismatchException Thrown if the number of time
     * intervals used to represent the user's power consumption is not the same
     * as the number of intervals used by this coalition.
     *
     * Adds a new user at the end of this coalition. Allows one to use this
     * class as a stack of users
     */
    public final void push(PowerUser u) throws TimeIntervalsNumberMismatchException {

        //Is there room for the new user?
        if (usersNumber >= maxUsersNumber) {

            //No room. Do nothing
            return;
        }

        if (u.timeIntervalNumber() != intervals) {

            //Error: the number of time intervals must match
            throw new TimeIntervalsNumberMismatchException();
        }

        //Add the user (shallow copy)
        users[usersNumber] = u;

        //Update the number of users
        ++usersNumber;

        //Create a local variable to store the intermediate sums
        float val;

        //Create an accumulator to find the new peak cumulative power consumption
        float max = Float.NEGATIVE_INFINITY;

        //Extract the power usage for this user
        float[] usage = u.getUsages();

        for (int i = 0; i < intervals; ++i) {

            //Get the old cumulative value and add the power usage at the
            //current time for the user being added
            val = cumulative[usersNumber - 1][i] + usage[i];

            //Update the max power consumption
            if (val > max) {

                max = val;
            }

            //Save the new cumulative consumption value.
            cumulative[usersNumber][i] = val;
        }

        //Save the new maximum power consumption
        currentMaxConsumption[usersNumber] = max;

        //Update the sum of maximums
        sumOfMaximums[usersNumber] = sumOfMaximums[usersNumber - 1] +
                                        u.getMaxConsumption();
    }

    /**
     * Removes the last user present in this coalition
     *
     * @throws EmptyStackException Thrown if the coalition represented by the
     * object on which this method is called is empty.
     *
     * Removes the last user in this coalition, allowing one to use this class
     * as a stack.
     */
    public final void pop() throws EmptyStackException {

        //Do not pop an empty stack
        if (usersNumber == 0) {
            
            //Throw an exception
            throw new EmptyStackException();
        }

        //Reduce the number of users present
        --usersNumber;
    }

    /**
     * Returns a reference to the last user in this coalition
     *
     * @return A reference to the last user in this coalition
     *
     * @throws EmptyStackException Thrown if the coalition represented by the
     * object on which this method is called is empty.
     *
     * This method allows access to a reference to the last user in this
     * coalition
     */
    public final PowerUser peek() throws EmptyStackException {

        //Make sure that the stack is not empty
        if (usersNumber > 0) {

            //Return the requested element
            return users[usersNumber - 1];
        }
        else {

            //Throw an exception
            throw new EmptyStackException();
        }
    }
}
