/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package electro.entities;

import java.awt.Color;
import java.util.ArrayList;

/**
 * This class represents a power user, that is a power utilizer
 *
 * @author Andrea Zambon
 */
public class PowerUser {

    private static int counter = 0;
    private static final Object mutex = new Object();

    private float[] powerUsage;
    private Color col;
    private final int intervals;
    private int userId;
    private int maxConsumptionTime;
    private int minConsumptionTime;

    /**
     * Creates a new PowerUsers from the specified power consumption data
     *
     * @param newUserPowerUsage The power consumption data for this user
     *
     * @param c The color that will be used to represent this user
     *
     * This constructor creates a new PowerUser object having the specified
     * power consumption and color, and with a progressive serial number assigned
     * to it.
     */
    public PowerUser(ArrayList<Float> newUserPowerUsage, Color c) {

        //Save the number of time intervals
        intervals = newUserPowerUsage.size();

        //Create the array of values for the power consumption
        powerUsage = new float[intervals];
        
        //Start with the initial time for the maximum and minimum power consumption
        maxConsumptionTime = -1;
        minConsumptionTime = -1;
        
        //Initialize a temporary variable to contain the maximum and minimum
        //power consumption
        float maxConsumption = Float.NEGATIVE_INFINITY;
        float minConsumption = Float.POSITIVE_INFINITY;

        //Copy value by value
        for (int i = 0; i < intervals; ++i) {

            powerUsage[i] = newUserPowerUsage.get(i).floatValue();

            //Update the index of the maximum value
            if (powerUsage[i] > maxConsumption) {

                maxConsumption = powerUsage[i];
                maxConsumptionTime = i;
            }

            //Same for the minimum. Do not convert to an else-if statement since
            //in the first execution of the loop both if statements will be
            //executed
            if (powerUsage[i] < minConsumption) {

                minConsumption = powerUsage[i];
                minConsumptionTime = i;
            }
        }

        //Save the color for this user
        col = c;

        //Execute the following insructions in mutual exclusion
        synchronized (mutex) {

            //Get an user id
            userId = counter;

            //Prepare the next user id
            ++counter;
        }
    }

    /**
     * Creates a new PowerUsers from the specified power consumption data
     *
     * @param newUserPowerUsage The power consumption data for this user
     *
     * @param c The color that will be used to represent this user
     *
     * This constructor creates a new PowerUser object having the specified
     * power consumption and color, and with a progressive serial number assigned
     * to it.
     */
    public PowerUser(float[] newUserPowerUsage, Color c) {

        //Save the number of time intervals
        intervals = newUserPowerUsage.length;

        //Create the array of values for the power consumption
        powerUsage = new float[intervals];

        //Start with the initial time for the maximum and minimum power consumption
        maxConsumptionTime = -1;
        minConsumptionTime = -1;

        //Initialize a temporary variable to contain the maximum and minimum
        //power consumption
        float maxConsumption = Float.NEGATIVE_INFINITY;
        float minConsumption = Float.POSITIVE_INFINITY;

        //Copy value by value
        for (int i = 0; i < intervals; ++i) {

            powerUsage[i] = newUserPowerUsage[i];

            //Update the index of the maximum value
            if (powerUsage[i] > maxConsumption) {

                maxConsumption = powerUsage[i];
                maxConsumptionTime = i;
            }

            //Same for the minimum
            if (powerUsage[i] < minConsumption) {

                minConsumption = powerUsage[i];
                minConsumptionTime = i;
            }
        }

        //Save the color for this user
        col = c;

        //Execute the following insructions in mutual exclusion
        synchronized (mutex) {

            //Get an user id
            userId = counter;

            //Prepare the next user id
            ++counter;
        }
    }

    /**
     * Gets the id of this user
     *
     * @return The id of this user
     *
     * This method retuns the id of this user. The id of an user is assigned to
     * it when it is created. It is generated using a static counter that is
     * incremented every time a PowerUser object is created.
     */
    public int getUserId() {

        return userId;
    }

    /**
     * Returns the number of time intervals present in the power
     * consumption profile of this user.
     *
     * @return The number of time intervals.
     *
     * This method returns the number of time intervals that represents the
     * power consumption profile for this user.
     */
    public int timeIntervalNumber() {

        //Number of time intervals
        return intervals;
    }

    /**
     * Returns the power consumption value at the specified time
     *
     * @param time The time at which the power consumption value is to be
     * returned
     *
     * @return The power consumption value at the specified time
     * 
     * @throws ArrayIndexOutOfBoundsException Thrown if the supplied time is
     * not valid
     *
     * This method allows one to access the power consumption values for every
     * time interval.
     */
    public final float usageAtTime(int time) {

        //Make sure that the index is valid
        if (time < 0 || time >= intervals) {
            
            throw new IndexOutOfBoundsException();
        }

        //Get the required value
        return powerUsage[time];
    }

    /**
     * Returns a reference to the internal array containing the power usage data
     * for this user
     *
     * @return A reference to an array of floats
     *
     * This method is used to quickly access the internal array of power
     * consumption values.
     */
    protected float[] getUsages() {

        return powerUsage;
    }

    /**
     * Compares the power consumptions of two users
     *
     * @param u The user whose power consumption is to be compared with the one
     * of this user.
     *
     * @return -1 if the power consumption of this user is samller than the one
     * of the other user, 0 if they are equal 1 otherwise.
     *
     * This method compares the power consumption of this user with the one of
     * another user.
     */
    public int compareValueTo(PowerUser u) {

        //Make sure that the number of time intervals is the same
        if (intervals != u.intervals) {

            //Different number of time samples
            if (intervals < u.intervals) {

                //Report that this is smaller
                return -1;
            }
            else {

                //Report that the other user is smaller
                return 1;
            }
        }

        //Compare time by time
        for (int i = 0; i < intervals; ++i) {

            if (powerUsage[i] < u.powerUsage[i]) {

                //This is smaller
                return -1;
            }
            else {

                if (powerUsage[i] > u.powerUsage[i]) {

                    //The other is smaller
                    return 1;
                }
            }
        }

        //Equal
        return 0;
    }

    /**
     * Sets the color for this user
     *
     * @param c The new color of this user
     *
     * This method allows to change the color of this user
     */
    public void setColor(Color c) {

        //Set the color
        col = c;
    }

    /**
     * This method allows one to obtain the color of this user
     *
     * @return The color of this user
     *
     * This method returns the color that is currently associated with this user.
     */
    public Color getColor() {

        //Return the color
        return col;
    }

    /**
     * Returns the maximum power consumption value
     *
     * @return The maximum power consumption value for this user.
     *
     * This method returns the maximum power consumption value for this user.
     */
    public final float getMaxConsumption() {

        return powerUsage[maxConsumptionTime];
    }

    /**
     * Returns the minimum power consumption value
     *
     * @return The minimum power consumption value for this user.
     *
     * This method returns the minimum power consumption value for this user.
     */
    public final float getMinConsumption() {

        return powerUsage[minConsumptionTime];
    }

    /**
     * Creates a deep copy of this user
     *
     * @return A deep copy of this user
     *
     * This method Performs a deep copy of this user. It copies the power
     * consumption profile, the color and also the user id (which will be the
     * same as the one of the invocation user).
     */
    @Override
    public PowerUser clone() {
        
        //Create the result variable
        PowerUser res;
        
        //Execute in mutual exclusion to prevent problems with the static counter
        //variable
        synchronized (mutex) {
        
            //Create a deep copy of this object
            res = new PowerUser(powerUsage, col);

            //Make sure that the user id is the same
            res.userId = userId;

            //Reset the counter to the previous value
            --counter;
        }

        //Return the result
        return res;
    }

    /**
     * This method resets the user numbering sequence
     *
     * This method allows one to make sure that the users created hereafter will
     * have their serial number forming a new sequence starting from 0
     */
    public static void resetUserNumbering() {

        counter = 0;
    }
}
