/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package electro.file;

import electro.entities.PowerUser;
import java.awt.Color;
import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.BitSet;

/**
 * This class handles all the CSV file read operations of the program
 *
 * @author Andrea Zambon
 */
public class CSVFileReader {

    private FileReader fr;
    private BufferedReader br;

    /**
     * Creates a new CSVFileReader object
     *
     * @param filename The file to open
     *
     * @throws FileNotFoundException Thrown if the specified file cannot be found
     *
     * This constructor creates a new CSVFileReader object to read the specified
     * file.
     */
    public CSVFileReader(String filename)
            throws FileNotFoundException {

        //Initialize the internal objects to a known state
        fr = null;
        br = null;

        //Try to open the file
        openFile(filename);
    }

    /**
     * Opens a file
     *
     * @param filename The file to open
     *
     * @throws FileNotFoundException Thrown if the specified file could not be
     * found
     *
     * This method tries to open the specified file.
     */
    public final void openFile(String filename) throws FileNotFoundException {

        //Make sure that no file is already open
        if (fr != null || br != null) {

            //Close the open file
            close();
        }

        //Try to open the new file
        fr = new FileReader(filename);
        br = new BufferedReader(fr);
    }

    /**
     * This method reads the specified number of users from an open file
     *
     * @param maxUsersNumber The maximum numbers of users to read. If the file
     * does not contain enough users to reach this number, this method reads all
     * the users in the file and then returns.
     *
     * @return An ArrayList of PowerUsers object representing the users read
     * 
     * @throws IOException If an error occurs while reading the file or if this
     * method is called when no file is open.
     *
     * This method tries to read the data for a specified number of users from
     * an open file. It also discards the data sets where there are zero values
     * and those where there are large jumps between two consecutive values. It
     * reads the serial number of each user and then tries to find a data set
     * that is valid. As soon as one is found for a certain user, that data set
     * is used to create a new user for the result and the algorithm moves to
     * the following user. Although many daily data sets may be available for
     * each users, only the first valid one is used.
     */
    public ArrayList<PowerUser> read(int maxUsersNumber) throws IOException {

        //Make sure that a file is open
        if (br == null) {

            //No open file. Throw an exceptions
            throw new IOException();
        }

        //Reset the user numbering to make sure that the new users have their
        //id numbers starting from 0
        PowerUser.resetUserNumbering();

        //Read the first line
        String line = br.readLine();

        //Create an array to contain the column headers
        ArrayList<String> colNames = new ArrayList<String>();

        //Index of the forst valid character after the last comma
        int firstValidCharacterAfterLastComma = 0;

        //Go until the end of the line is reached
        for (int i = 0; i < line.length(); ++i) {

            //Look for a comma
            if (line.charAt(i) == ',') {

                //Extract the substring and add it to the array
                colNames.add(line.substring(firstValidCharacterAfterLastComma, i));

                //Save the index of the character after this comma
                firstValidCharacterAfterLastComma = i + 1;
            }
        }

        //Add the last item
        colNames.add(line.substring(firstValidCharacterAfterLastComma, line.length()));

        //Find the user id column
        int userIdColumn = colNames.indexOf("Meter Serial No");

        //Create a mask of bits to identify data columns
        BitSet dataColumns = new BitSet();

        //Check all column headers to find data columns
        for (int i = 0; i < colNames.size(); ++i) {

            if (colNames.get(i).contains("DIP")) {

                //Set the corresponding bit
                dataColumns.set(i);
            }
        }

        //Create an array of rows (arrays of strings) to contain the rest
        //of the file
        ArrayList<ArrayList<String>> values = new ArrayList<ArrayList<String>>();

        //Current row index
        int row = 0;

        while (true) {

            //Reset the index of the first character after the last comma
            firstValidCharacterAfterLastComma = 0;

            //Read a line
            line = br.readLine();

            if (line == null) {

                //Reached the end of the file
                break;
            }

            //Add a new row
            values.add(new ArrayList<String>());

            for (int i = 0; i < line.length(); ++i) {

                //Look for a comma
                if (line.charAt(i) == ',') {

                    //Add the substring since the last comma to the end of the
                    //current row
                    values.get(row).add(line.substring(firstValidCharacterAfterLastComma, i));

                    //Set the first valid character to the one after this one
                    firstValidCharacterAfterLastComma = i + 1;
                }
            }

            //Add the last element of this row
            values.get(row).add(line.substring(firstValidCharacterAfterLastComma,
                                                line.length()));

            //Move to the next row
            ++row;

        }   //while

        //Save the total number of rows
        int totalRows = row;

        //Start over from the first row
        row = 0;

        //Initially set the user id to an invalid value
        String lastUserId = "";

        //Initially set this flag to represent that a previous user has already
        //been processed
        boolean newId = false;

        //Create an array for the result
        ArrayList<PowerUser> users = new ArrayList<PowerUser>();

        //Initialize a counter for the total number of users read
        int usersNumber = 0;

        //Iterate on all rows
        while (row < totalRows) {

            //New user id?
            if (values.get(row).get(userIdColumn).compareTo(lastUserId) != 0) {
                
                //New id
                newId = true;
                lastUserId = values.get(row).get(userIdColumn);
            }

            if (newId) {

                //Are these values valid?
                //Initially there is no previous value
                float lastVal = -1;
                float val = -1;

                //Initially assume that the values are valid
                boolean valid = true;

                for (int i = 0; i < colNames.size(); ++i) {

                    //Only check data columns
                    if (dataColumns.get(i)) {

                        //Update the last value
                        lastVal = val;

                        //Get the next value
                        val = new Float(values.get(row).get(i));

                        //Is this the first value?
                        if (lastVal != -1) {

                            //Not the first value. Check that the new value is
                            //not negative or too different from the last one
                            if (val <= 0 ||
                                    java.lang.Math.abs((val - lastVal) / val) > 5) {

                                //Invalid
                                valid = false;
                                break;
                            }
                        }
                        else {

                            //First value. Only check that it is positive
                            if (val <= 0) {

                                //Invalid
                                valid = false;
                                break;
                            }
                        }
                    }
                }

                //Is this a valid row?
                if (valid) {

                    //Add this line to the result. First must convert all string
                    //values to floats
                    ArrayList<Float> user = new ArrayList<Float>();

                    //Read every data value
                    for (int i = 0; i < colNames.size(); ++i) {

                        //Only convert data columns
                        if (dataColumns.get(i)) {

                            //Convert and save the result
                            user.add(new Float(values.get(row).get(i)));
                        }
                    }

                    //Get a color using incrementally different hues with maximum
                    //saturation and brightness
                    Color c = Color.getHSBColor((usersNumber / (float)maxUsersNumber),
                                                    1.0f, 1.0f);

                    //Create a new object with the new data and add it to the
                    //result
                    users.add(new PowerUser(user, c));

                    //Increment the number of users read
                    ++usersNumber;

                    //Did we read all users allowed?
                    if (usersNumber >= maxUsersNumber) {

                        //Early termination
                        return users;
                    }

                    //Remember that a row for this user id has already been found
                    newId = false;
                }
            }

            //Move to the next row
            ++row;
        }

        //Done
        return users;
    }

    /**
     * Closes an open file
     *
     * This method tries to close a file that is open.
     */
    public void close() {

        try {

            //Close the file
            br.close();
            fr.close();
        }

        catch (IOException ex) {

            //Ignore the exception
        }

        //Remove every reference to the file in this object
        br = null;
        fr = null;
    }
}
